"""Python Toodledo API

== MIT License ==

Copyright (c) 2011 Erik Demaine

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import datetime, hashlib, json, os, sys, time, urllib
import cPickle as pickle
pickle_protocol = pickle.HIGHEST_PROTOCOL
import bsddb.db as bsddb

token_url = 'https://api.toodledo.com/2/account/token.php'
lookup_url = 'https://api.toodledo.com/2/account/lookup.php'
## These need Pro to work with https:
account_info_url = 'http://api.toodledo.com/2/account/get.php'
tasks_get_url = 'http://api.toodledo.com/2/tasks/get.php'
tasks_deleted_url = 'http://api.toodledo.com/2/tasks/deleted.php'
folders_get_url = 'http://api.toodledo.com/2/folders/get.php'
contexts_get_url = 'http://api.toodledo.com/2/contexts/get.php'
goals_get_url = 'http://api.toodledo.com/2/goals/get.php'
locations_get_url = 'http://api.toodledo.com/2/locations/get.php'

## Actual expiration is 4 hours.  Expire 5 minutes early to avoid timing issues.
expiration = datetime.timedelta (hours = 4, minutes = -5)

## These don't include the always-returned id, title, modified, completed
nonmeta_fields = ['folder', 'context', 'goal', 'location', 'tag', 'startdate', 'duedate', 'duedatemod', 'starttime', 'duetime', 'remind', 'repeat', 'status', 'star', 'priority', 'length', 'timer', 'added', 'note', 'parent', 'children', 'order']
all_fields = nonmeta_fields + ['meta']

def md5 (x):
  return hashlib.md5 (x).hexdigest ()

def toodledo_query (url, query):
  data = json.load (urllib.urlopen (url, urllib.urlencode (query)))
  if 'errorDesc' in data:
    raise RuntimeError ('Toodledo: ' + data['errorDesc'])
  return data

class Toodledo:
  appid = None       ## must be overriden or specified in __init__
  apptoken = None    ## must be overriden or specified in __init__
  appversion = None  ## can be overriden or specified in __init__
  device = '%s Python %s' % (sys.platform, sys.version.split()[0])
  osver = None  ## must be an integer

  def __init__ (self, dbdir, email, password,
                appid = None, apptoken = None, appversion = None,
                device = None, osver = None):
    self.dbdir = str (dbdir)
    self.email = str (email)
    self.password = str (password)
    if appid is not None: self.appid = appid
    if apptoken is not None: self.apptoken = apptoken
    if appversion is not None: self.appversion = appversion
    if device is not None: self.device = device
    if osver is not None: self.osver = osver
    self._userid = None

    if not os.path.isdir (self.dbdir):
      os.mkdir (self.dbdir)
    self.dbenv = bsddb.DBEnv ()
    self.dbenv.open (self.dbdir,
      bsddb.DB_INIT_LOCK | bsddb.DB_INIT_MPOOL | bsddb.DB_CREATE)
    self.useriddb = bsddb.DB (self.dbenv)
    self.useriddb.open ('userid.db', None, bsddb.DB_HASH, bsddb.DB_CREATE)
    self.tokendb = bsddb.DB (self.dbenv)
    self.tokendb.open ('token.db', None, bsddb.DB_HASH, bsddb.DB_CREATE)
    self.lastcheck = bsddb.DB (self.dbenv)
    self.lastcheck.open ('last.db', None, bsddb.DB_HASH, bsddb.DB_CREATE)

  def userid (self):
    if self._userid is not None: return self._userid
    data = self.useriddb.get (self.email)
    if data is not None:
      return data
    query = [
      ('appid', self.appid), ('email', self.email), ('pass', self.password),
      ('sig', md5 (self.email + self.apptoken))
    ]
    data = toodledo_query (lookup_url, query)
    userid = str (data['userid'])
    self.useriddb.put (self.email, userid)
    self._userid = userid
    return userid

  def db_key (self):
    if self.appid is None or self.apptoken is None:
      raise RuntimeError ('Toodledo appid/apptoken unspecified')
    return self.appid + '\0' + self.userid ()

  def session_token (self):
    data = self.tokendb.get (self.db_key ())
    if data is not None:
      data = pickle.loads (data)
      if data['expires'] > datetime.datetime.utcnow ():
        return data['token']
    sig = md5 (self.userid () + self.apptoken)
    query = [(x,y) for x,y in
      [('userid', self.userid ()), ('appid', self.appid),
       ('vers', self.appversion), ('device', self.device), ('os', self.osver),
       ('sig', sig)]
      if y is not None]
    data = toodledo_query (token_url, query)
    token = str (data['token'])
    self.tokendb.put (self.db_key (), pickle.dumps ({
      'expires': datetime.datetime.utcnow () + expiration,
      'token': token
    }))
    return token

  def session_key (self):
    return md5 (md5 (self.password) + self.apptoken + self.session_token ())

  def account_info (self):
    query = [('key', self.session_key ())]
    data = toodledo_query (account_info_url, query)
    return data

  def tasks (self, modbefore = None, modafter = None, completed = None,
             fields = all_fields):
    query = [('key', self.session_key ())]
    if modbefore is not None:
      query.append (('modbefore', modbefore))
      #time.mktime (modbefore.utctimetuple ())))
    if modafter is not None:
      query.append (('modafter', modafter))
      #time.mktime (modafter.utctimetuple ())))
    if completed:
      query.append (('completed', 1))
    elif completed is not None:
      query.append (('completed', 0))
    query.append (('fields', ','.join (fields)))
    query.append (('num', 1000000))
    data = toodledo_query (tasks_get_url, query)
    return data[1:]

  def deleted_tasks (self, after = None):
    query = [('key', self.session_key ())]
    if after is not None:
      query.append (('after', after))
      #time.mktime (after.utctimetuple ())))
    data = toodledo_query (tasks_deleted_url, query)
    return data[1:]

  def folders (self, after = None):
    query = [('key', self.session_key ())]
    data = toodledo_query (folders_get_url, query)
    return data

  def contexts (self, after = None):
    query = [('key', self.session_key ())]
    data = toodledo_query (contexts_get_url, query)
    return data

  def goals (self, after = None):
    query = [('key', self.session_key ())]
    data = toodledo_query (goals_get_url, query)
    return data

  def locations (self, after = None):
    query = [('key', self.session_key ())]
    data = toodledo_query (locations_get_url, query)
    return data

  TYPES = ['tasks', 'deleted_tasks', 'folders', 'contexts', 'goals', 'locations']

  def updates (self, tasks = True, deleted_tasks = True, folders = True, contexts = True, goals = True, locations = True):
#lastedit_task : A timestamp that indicates the last time that any task was added or edited on this account. You can quickly check this field to determine if you need to download updates. lastdelete_task : A timestamp that indicates the last time that any task was deleted from this account. You can quickly check this field to determine if you need to identify and remove tasks from your application. lastedit_folder : A timestamp that indicates the last time that a folder was added, edited or deleted. You can quickly check this field to determine if you need to refresh your cached folder list. lastedit_context : A timestamp that indicates the last time that a context was added, edited or deleted. You can quickly check this field to determine if you need to refresh your cached context list. lastedit_goal : A timestamp that indicates the last time that a goal was added, edited or deleted. You can quickly check this field to determine if you need to refresh your cached goal list. lastedit_location : A timestamp that indicates the last time that a location was added, edited or deleted. You can quickly check this field to determine if you need to refresh your cached location list. lastedit_notebook : A timestamp that indicates the last time that any notebook was added or edited on this account. You can quickly check this field to determine if you need to download updates. lastdelete_notebook : A timestamp that indicates the last time that any notebook was d
    lastcheck = self.lastcheck.get (self.db_key ())
    if lastcheck is None:
      lastcheck = {}
    else:
      lastcheck = pickle.loads (lastcheck)
    now = dict ((x[4:],y) for x,y in self.account_info ().iteritems ()
                          if x.startswith ('last'))
    out = {}
    if tasks and now.get ('edit_task') != lastcheck.get ('edit_task'):
      out['tasks'] = self.tasks (modafter = lastcheck.get ('edit_task'))
      lastcheck['edit_task'] = now.get ('edit_task')
    if deleted_tasks and now.get ('delete_task') != lastcheck.get ('delete_task'):
      out['deleted_tasks'] = self.deleted_tasks (after = lastcheck.get ('delete_task'))
      lastcheck['delete_task'] = now.get ('delete_task')
    if folders and now.get ('edit_folder') != lastcheck.get ('edit_folder'):
      out['folders'] = self.folders ()
      lastcheck['edit_folder'] = now.get ('edit_folder')
    if contexts and now.get ('edit_context') != lastcheck.get ('edit_context'):
      out['contexts'] = self.contexts ()
      lastcheck['edit_context'] = now.get ('edit_context')
    if goals and now.get ('edit_goal') != lastcheck.get ('edit_goal'):
      out['goals'] = self.goals ()
      lastcheck['edit_goal'] = now.get ('edit_goal')
    if locations and now.get ('edit_location') != lastcheck.get ('edit_location'):
      out['locations'] = self.locations ()
      lastcheck['edit_location'] = now.get ('edit_location')
    self.lastcheck.put (self.db_key (), pickle.dumps (lastcheck))
    return out
